//
//  LADatabaseManager.h
//  LAAdditions
//
//  Created by Cyril Anger on 14/01/2015.
//  Copyright (c) 2015 LyricApps. All rights reserved.
//

#import <Foundation/Foundation.h>

#import <LAAdditions/LADuration.h>
#import <LAAdditions/LAConstant.h>
#import <LAAdditions/LACountedSet.h>
#import <LAAdditions/LARequestsMap.h>
#import <LAAdditions/LADatabaseOperation.h>
#import <LAAdditions/LADocumentSynchronizationHandler.h>

#if TARGET_OS_IPHONE
@class LAMDocument;
#else
@class LADocument;
#endif
@class LADatabaseObject;
@class LAFiltersBarData;
@class LADatabaseSynchronizationSession;
@protocol LADatabaseSynchronizationSessionController;

extern NSString *LADatabaseErrorDomain;

typedef enum {
	LADatabaseErrorNone = 0,
	LADatabaseErrorOpenFailed,
	LADatabaseErrorInvalidQuery,
	LADatabaseErrorQueryExecutionFailed,
	LADatabaseErrorSaveFailed
} LADatabaseErrorCode;

@interface LADatabaseQuery : NSObject

- (void)reset;
- (void)setNullForParameter:(NSString *)parameter;
- (void)setBool:(BOOL)value forParameter:(NSString *)parameter;
- (void)setInt:(int)value forParameter:(NSString *)parameter;
- (void)setInt64:(long long)value forParameter:(NSString *)parameter;
- (void)setDouble:(double)value forParameter:(NSString *)parameter;
- (void)setString:(NSString *)value forParameter:(NSString *)parameter;
- (void)setData:(NSData *)value forParameter:(NSString *)parameter;
- (void)setDate:(NSDate *)value forParameter:(NSString *)parameter;
- (void)setDateTime:(NSDate *)value forParameter:(NSString *)parameter;
- (void)setDuration:(LADuration *)value forParameter:(NSString *)parameter;
- (void)setDecimalNumber:(NSDecimalNumber *)value forParameter:(NSString *)parameter;
- (void)setImage:(LAImage *)image forParameter:(NSString *)parameter;
- (void)setColor:(LAColor *)color forParameter:(NSString *)parameter;
- (void)setJSONObject:(id)object forParameter:(NSString *)parameter;
- (void)setConstant:(LAConstant *)constant forParameter:(NSString *)parameter;
- (void)setClass:(Class)class forParameter:(NSString *)parameter;

@end

typedef void (^LADatabaseChangePrepareBlock)(LADatabaseQuery *query);

@interface LADatabaseChange : NSObject

@property (weak, nonatomic) LADatabaseObject *object;

- (instancetype)initWithSQL:(NSString *)sql prepareBlock:(LADatabaseChangePrepareBlock)prepareBlock;

@end

@interface LADatabaseResultSet : NSObject

- (BOOL)next;
- (BOOL)isNullForColumn:(NSString *)column;
- (BOOL)boolForColumn:(NSString *)column;
- (int)intForColumn:(NSString *)column;
- (long long)int64ForColumn:(NSString *)column;
- (double)doubleForColumn:(NSString *)column;
- (NSString *)stringForColumn:(NSString *)column;
- (NSData *)dataForColumn:(NSString *)column;
- (NSDate *)dateForColumn:(NSString *)column;
- (NSDate *)dateTimeForColumn:(NSString *)column;
- (LADuration *)durationForColumn:(NSString *)column;
- (NSDecimalNumber *)decimalNumberForColumn:(NSString *)column;
- (LAImage *)imageForColumn:(NSString *)column;
- (LAColor *)colorForColumn:(NSString *)column;
- (id)JSONObjectForColumn:(NSString *)column;
- (id)constantForColumn:(NSString *)column;
- (Class)classForColumn:(NSString *)column;

@end

extern NSString *LADatabaseManagerObjectsDidChangeNotification;

@interface LADatabaseManager : NSObject

- (instancetype)init NS_UNAVAILABLE;
- (instancetype)initWithURL:(NSURL *)URL document:(id)document readOnly:(BOOL)readOnly deleteOnDealloc:(BOOL)deleteOnDealloc error:(NSError * __autoreleasing *)error NS_DESIGNATED_INITIALIZER;
@property (strong, nonatomic, readonly) NSURL *URL;
#if TARGET_OS_IPHONE
@property (weak, nonatomic, readonly) LAMDocument *document;
#else
@property (weak, nonatomic, readonly) LADocument *document;
#endif
@property (getter = isLoaded, nonatomic, readonly) BOOL loaded;
@property (getter = isReadOnly, nonatomic, readonly) BOOL readOnly;
@property (nonatomic, readonly) BOOL deleteOnDealloc;
@property (strong, nonatomic, readonly) NSUndoManager *undoManager;

#pragma mark Copy

- (instancetype)initByCopyingDatabaseManager:(id)databaseManager NS_DESIGNATED_INITIALIZER;
@property (weak, nonatomic, readonly) LADatabaseManager *originalDatabaseManager;
- (instancetype)copy;

#pragma mark Database

+ (NSString *)version;
+ (NSString *)createStatementsFile;
+ (NSArray *)updateStatementsFilesForVersion:(NSString *)version;
- (LADatabaseQuery *)prepareQuery:(NSString *)sql error:(NSError * __autoreleasing *)error;
- (LADatabaseResultSet *)executeQuery:(LADatabaseQuery *)query error:(NSError * __autoreleasing *)error;
- (BOOL)executeUpdate:(LADatabaseQuery *)query error:(NSError * __autoreleasing *)error;
- (BOOL)executeCustomUpdateStatementsForVersion:(NSString *)version error:(NSError * __autoreleasing *)error;

#pragma mark Objects

+ (Class)rootObjectClass;
@property (strong, nonatomic, readonly) LADatabaseObject *rootObject;
- (void)addInitialData;
- (void)addSampleData;
- (id)objectWithID:(NSString *)ID;
- (NSArray *)objectsWithIDs:(NSArray *)IDs;
- (void)registerObject:(LADatabaseObject *)object;
- (void)unregisterObject:(LADatabaseObject *)object;
- (void)requestObjectWithID:(NSString *)ID usingBlock:(LARequestBlock)block;
- (void)removeAllObjectRequests;
@property (strong, nonatomic, readonly) NSMutableDictionary *deletedIDs;
- (BOOL)isObjectDeleted:(NSString *)ID;
- (void)addDeletedID:(NSString *)ID;
- (void)addDeletedID:(NSString *)ID date:(NSDate *)date;
- (void)removeDeletedID:(NSString *)ID;
@property (nonatomic, readonly) BOOL shouldModifyLastModificationDates;
- (void)performBlockWithoutModifyingLastModificationDates:(void (^)(void))block;
@property (strong, nonatomic, readonly) NSMutableDictionary *filtersBarDatas;
- (LAFiltersBarData *)filtersBarData:(NSString *)name;
- (LACountedSet *)completionsSet:(NSString *)keyPath;

#pragma mark Metadata

@property (strong, nonatomic, readonly) NSString *password;
- (BOOL)checkPassword:(NSString *)password;
- (void)protectWithPassword:(NSString *)password;
- (id)valueForLocalParameter:(NSString *)parameter;
- (void)setValue:(id)value forLocalParameter:(NSString *)parameter;

#pragma mark Data

- (void)loadData;
- (void)dataDidLoad;
- (void)prepareForImportingOldData;
- (void)finishImportingOldData:(BOOL)success;
- (NSArray *)computeChanges;
- (BOOL)saveChanges:(NSArray *)changes error:(NSError * __autoreleasing *)error;
- (void)commitChangesToOriginalDatabaseManager;

#pragma mark Operations

- (LADatabaseOperation *)addOperationWithBlock:(LADatabaseOperationBlock)block;

#pragma mark Synchronization

@property (strong, nonatomic, readonly) NSArray *synchronizationHandlers;
- (LADocumentSynchronizationHandler *)synchronizationHandler:(NSString *)description;
- (NSData *)saveChanges:(NSArray *)changes processSynchronizationHandlers:(BOOL)processSynchronizationHandlers displayName:(NSString *)displayName error:(NSError * __autoreleasing *)error;
- (BOOL)saveChanges:(NSArray *)changes andCopyToURL:(NSURL *)URL processSynchronizationHandlers:(BOOL)processSynchronizationHandlers error:(NSError * __autoreleasing *)error;
- (void)synchronizeWithDatabaseAtURL:(NSURL *)URL inBackground:(BOOL)inBackground otherIcon:(LAImage *)otherIcon startHandler:(void (^)(id <LADatabaseSynchronizationSessionController> synchronizationSessionController))startHandler completionHandler:(void (^)(BOOL canceled, NSString *error))completionHandler;
- (void)cancelPendingSynchronizations;
+ (NSUInteger)numberOfSynchronizationSteps;
- (void)prepareSynchronization:(NSUInteger)step withDatabaseManager:(LADatabaseManager *)databaseManager;
+ (NSUInteger)maximumSynchronizationPriorityForStep:(NSUInteger)step;
- (NSArray *)synchronizingObjectIDsForStep:(NSUInteger)step;

#pragma mark Duplication

- (NSData *)archivedDataWithRootObject:(id)object;
- (id)unarchiveObjectWithData:(NSData *)data;

@end

@protocol LADatabaseManagerDocument <NSObject>

- (id <LADatabaseSynchronizationSessionController>)startSynchronizationSessionController:(LADatabaseSynchronizationSession *)synchronizationSession inBackground:(BOOL)inBackground withCompletionHandler:(void (^)(BOOL canceled, NSString *error))completionHandler;

@end

@interface NSCoder (LADatabaseManagerAdditions)

@property (weak, nonatomic) LADatabaseManager *databaseManager;

@end
